from django.db import models
from django.contrib.auth.models import User
from mongoengine import Document, StringField, IntField, DecimalField, ListField, BooleanField, DictField
from django.db.models.signals import post_init
import json
import datetime

nationalities = ["Norwegian",
                 "Other Nordic",
                 "United Kingdom and Ireland",
                 "Germany and Benelux",
                 "Southern and Western Europe",
                 "Eurasia",
                 "Central Europe",
                 "North America",
                 "Other America",
                 "China and North-West Asia",
                 "Other Asia",
                 "Africa"]

class RestaurantSimilarityMatrix(models.Model):
    restaurant = models.CharField(max_length=100, primary_key=True, unique=True)
    recs = models.TextField()

    def set_recs(self, x):
        self.recs = json.dumps(x)

    def get_recs(self):
        return json.loads(self.recs)

    def __str__(self):
        return self.restaurant


class UserSimilarityMatrix(models.Model):
    user = models.CharField(max_length=100)
    context = models.ForeignKey("Context", on_delete=models.CASCADE)
    recs = models.TextField()

    def set_recs(self, x):
        self.recs = json.dumps(x)

    def get_recs(self):
        return json.loads(self.recs)

    def __str__(self):
        return self.restaurant

    class Meta:
        unique_together = (('user', 'context'))


class ValidateRating(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    restaurant = models.CharField(max_length=100)
    context = models.ForeignKey("Context", on_delete=models.CASCADE)
    value = models.IntegerField(default=0)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    method = models.CharField(max_length=20)

    class Meta:
        unique_together = (('user', 'restaurant', 'context', 'method'))

    def __str__(self):
        return "{} rated restaurant {} {} in context {} at {} in origin {}".format(self.user, self.restaurant, self.value,
                                                                      self.context, self.updated_at, self.origin)

class UserProfile(models.Model):
    SEX_CHOICES = (('M', 'Male'), ('F', 'Female'),)
    NATIONALITY_CHOICES = [(x, x) for x in nationalities]
    YEAR_CHOICES = [(r, r) for r in range(1930, (datetime.datetime.now().year + 1))]

    sex = models.CharField(max_length=1, choices=SEX_CHOICES, blank=True)
    nationality = models.CharField(max_length=50, choices=NATIONALITY_CHOICES, blank=True, null=True)
    birthyear = models.IntegerField(choices=YEAR_CHOICES, blank=True, null=True)

    user = models.OneToOneField(User, on_delete=models.CASCADE)

    def get_sex_as_int(self):
        return 1.000 if self.sex == 'M' else 0.000

    def get_age_group_vector(self):
        this_year = datetime.datetime.now().year
        age = this_year - self.birthyear
        vector = [0.000] * 6

        if age < 20:
            vector[0] = 1.000
        elif age < 30:
            vector[1] = 1.000
        elif age < 40:
            vector[2] = 1.000
        elif age < 50:
            vector[3] = 1.000
        elif age < 60:
            vector[4] = 1.000
        else:
            vector[5] = 1.000
        return vector

    def get_nationality_group_vector(self):
        vector = [0.000] * len(nationalities)
        index = nationalities.index(self.nationality)
        vector[index] = 1.000
        return vector


class Context(models.Model):
    name = models.CharField(max_length=20, unique=True)

    def __str__(self):
        return self.name


class Cuisine(models.Model):
    name = models.CharField(primary_key=True, max_length=20, unique=True)

    def __str__(self):
        return self.name


class Restaurant(Document):
    factual_id = StringField(max_length=100, primary_key=True)
    name = StringField(max_length=100)
    address = StringField(max_length=100)
    address_extended = StringField(max_length=100)
    po_box = StringField(max_length=100)
    locality = StringField(max_length=100)
    neighborhood = ListField()
    region = StringField(max_length=100)
    postcode = StringField(max_length=100)
    country = StringField(max_length=100)
    latitude = DecimalField()
    longitude = DecimalField()

    tel = StringField(max_length=100)
    fax = StringField(max_length=100)
    website = StringField(max_length=500)
    email = StringField(max_length=100)
    owner = StringField(max_length=100)

    cuisine = ListField(StringField())
    price = IntField()
    rating = DecimalField()
    chain_id = StringField(max_length=100)
    chain_name = StringField(max_length=100)
    category_ids = ListField()
    category_labels = ListField()
    hours = DictField()
    hours_display = StringField(max_length=500)
    founded = StringField(max_length=100)
    attire = StringField(max_length=100)
    attire_required = ListField(StringField())
    attire_prohibited = ListField(StringField())
    admin_region = StringField(max_length=100)

    payment_cashonly = BooleanField()
    reservations = BooleanField()
    open_24hrs = BooleanField()
    parking = BooleanField()
    parking_valet = BooleanField()
    parking_garage = BooleanField()
    parking_street = BooleanField()
    parking_lot = BooleanField()
    parking_validated = BooleanField()
    parking_free = BooleanField()
    smoking = BooleanField()
    meal_breakfast = BooleanField()
    meal_lunch = BooleanField()
    meal_dinner = BooleanField()
    meal_deliver = BooleanField()
    meal_takeout = BooleanField()
    meal_cater = BooleanField()
    alcohol = BooleanField()
    alcohol_bar = BooleanField()
    alcohol_beer_wine = BooleanField()
    alcohol_byob = BooleanField()
    kids_goodfor = BooleanField()
    kids_menu = BooleanField()
    groups_goodfor = BooleanField()
    accessible_wheelchair = BooleanField()
    seating_outdoor = BooleanField()
    wifi = BooleanField()
    room_private = BooleanField()
    options_vegetarian = BooleanField()
    options_vegan = BooleanField()
    options_glutenfree = BooleanField()
    options_lowfat = BooleanField()
    options_organic = BooleanField()
    options_healthy = BooleanField()

    def get_vector(self):
        return self.get_cuisine_vector() + \
               self.get_price_vector() + \
               self.get_rating_vector() +\
               self.get_attire_vector() +\
            [self.get_reservations(),
             self.get_parking(),
             self.get_smoking(),
             self.get_alcohol(),
             self.get_wifi(),
             self.get_kids_menu(),
             self.get_options_vegetarian(),
             self.get_meal_takeout(),
             self.get_accessible_wheelchair(),
             self.get_seating_outdoor()]

    def get_boolean_vector(self):
        return [self.get_reservations(),
             self.get_parking(),
             self.get_smoking(),
             self.get_alcohol(),
             self.get_wifi(),
             self.get_kids_menu(),
             self.get_options_vegetarian(),
             self.get_meal_takeout(),
             self.get_accessible_wheelchair(),
             self.get_seating_outdoor()]

    def get_cuisine_string(self):
        return ", ".join(self.cuisine)

    def get_cuisine_vector(self):
        cuisines = [cuisine.name for cuisine in Cuisine.objects.all()]
        vector = [0.000] * len(cuisines)
        if self.cuisine:
            for cuisine in self.cuisine:
                index = cuisines.index(cuisine)
                vector[index] = 1.000
        return vector

    def get_rating_vector(self):
        ratings = [1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0]
        vector = [0.000] * len(ratings)
        if self.rating:
            index = ratings.index(self.rating)
            vector[index] = 1.000
        return vector

    def get_price_vector(self):
        vector = [0.000] * 5
        if self.price:
            vector[self.price - 1] = 1.000
        return vector

    def get_price_range(self):
        prices = {
            1: "$15",
            2: "$15-30",
            3: "$30-50",
            4: "$50-75",
            5: "$75+"
        }
        return prices[self.price]

    def get_attire_vector(self):
        attires = ["streetwear",
                   "casual",
                   "business casual",
                   "smart casual",
                   "formal"
                   ]
        vector = [0.000] * len(attires)
        if self.attire:
            index = attires.index(self.attire)
            vector[index] = 1.000
        return vector

    def get_reservations(self):
        return 1.000 if self.reservations else 0.000

    def get_parking(self):
        return 1.000 if self.parking else 0.000

    def get_smoking(self):
        return 1.000 if self.smoking else 0.000

    def get_alcohol(self):
        return 1.000 if self.alcohol else 0.000

    def get_wifi(self):
        return 1.000 if self.wifi else 0.000

    def get_kids_menu(self):
        return 1.000 if self.kids_menu else 0.000

    def get_options_vegetarian(self):
        return 1.000 if self.options_vegetarian else 0.000

    def get_meal_takeout(self):
        return 1.000 if self.meal_takeout else 0.000

    def get_accessible_wheelchair(self):
        return 1.000 if self.accessible_wheelchair else 0.000

    def get_seating_outdoor(self):
        return 1.000 if self.seating_outdoor else 0.000

    def __str__(self):
        return self.name


class UserRating(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    restaurant = models.CharField(max_length=100)
    context = models.ForeignKey("Context", on_delete=models.CASCADE)
    value = models.IntegerField(default=0)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('user', 'restaurant', 'context'))

    def __str__(self):
        return "{} rated restaurant {} {} in context {} at {}".format(self.user, self.restaurant, self.value,
                                                                      self.context, self.updated_at)


class UserVector(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    context = models.ForeignKey("Context", on_delete=models.CASCADE)
    default_cuisine = [0.500] * len(Cuisine.objects.all())

    cuisine = models.CharField(max_length=1000, default=json.dumps(default_cuisine))
    price = models.CharField(max_length=200, default='[0.5,0.5,0.5,0.5,0.5]')
    rating = models.CharField(max_length=200, default='[0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5]')
    attire = models.CharField(max_length=200, default='[0.5,0.5,0.5,0.5,0.5]')
    reservations = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    parking = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    smoking = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    alcohol = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    wifi = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    kids_menu = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    options_vegetarian = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    meal_takeout = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    accessible_wheelchair = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)
    seating_outdoor = models.DecimalField(default=0.500, max_digits=10, decimal_places=3)

    def set_cuisine(self, x):
        self.cuisine = json.dumps(x)

    def get_cuisine(self):
        return json.loads(self.cuisine)

    def set_price(self, x):
        self.price = json.dumps(x)

    def get_price(self):
        return json.loads(self.price)

    def set_rating(self, x):
        self.rating = json.dumps(x)

    def get_rating(self):
        return json.loads(self.rating)

    def set_attire(self, x):
        self.attire = json.dumps(x)

    def get_attire(self):
        return json.loads(self.attire)

    def get_vector(self):
        return [self.profile.get_sex_as_int(),
                self.profile.get_age_group_vector(),
                self.profile.get_nationality_group_vector(),
                self.get_cuisine(),
                self.get_price(),
                self.get_rating(),
                self.get_attire(),
                self.reservations,
                self.parking,
                self.smoking,
                self.alcohol,
                self.wifi,
                self.kids_menu,
                self.options_vegetarian,
                self.meal_takeout,
                self.accessible_wheelchair,
                self.seating_outdoor
                ]

    def get_vector_without_demo(self):
        return  self.get_cuisine() + \
                self.get_price() + \
                self.get_rating() + \
                self.get_attire() + \
                [float(self.reservations),
                float(self.parking),
                float(self.smoking),
                float(self.alcohol),
                float(self.wifi),
                float(self.kids_menu),
                float(self.options_vegetarian),
                float(self.meal_takeout),
                float(self.accessible_wheelchair),
                float(self.seating_outdoor)
                ]

    def get_demo_vector(self):
        return [self.profile.get_sex_as_int()] + self.profile.get_age_group_vector() + self.profile.get_nationality_group_vector()

    def update_val_from_list(self, old_value, rest_value, value):
        new_value = round(float(old_value) + ((0.03 * value) * rest_value),4)
        if new_value > 1.0:
            return 1.000
        elif new_value < 0.0:
            return 0.000
        else:
            return new_value

    def update_val(self, old_value, rest_value, value):
        rest_value = -1.000 if rest_value== 0.000 else 1.000
        new_value = round(float(old_value) + ((0.03 * value) * rest_value),4)
        if new_value > 1.0:
            return 1.000
        elif new_value < 0.0:
            return 0.000
        else:
            return new_value

    def update_field_vector(self, old_list, rest_list, value):
        new_list = []
        for i in range(len(old_list)):
            new_list.append(self.update_val_from_list(old_list[i],rest_list[i],value))
        return new_list

    def update(self, user_rating):
        rest = Restaurant.objects.get(factual_id=user_rating.restaurant)

        self.set_cuisine(self.update_field_vector(self.get_cuisine(), rest.get_cuisine_vector(), user_rating.value))
        self.set_price(self.update_field_vector(self.get_price(), rest.get_price_vector(), user_rating.value))
        self.set_rating(self.update_field_vector(self.get_rating(), rest.get_rating_vector(), user_rating.value))
        self.set_attire(self.update_field_vector(self.get_attire(), rest.get_attire_vector(), user_rating.value))

        self.reservations = self.update_val(self.reservations, rest.get_reservations(), user_rating.value)
        self.parking = self.update_val(self.parking, rest.get_parking(), user_rating.value)
        self.smoking = self.update_val(self.smoking, rest.get_smoking(), user_rating.value)
        self.alcohol = self.update_val(self.alcohol, rest.get_alcohol(), user_rating.value)
        self.wifi = self.update_val(self.wifi, rest.get_wifi(), user_rating.value)
        self.kids_menu = self.update_val(self.kids_menu, rest.get_kids_menu(), user_rating.value)
        self.options_vegetarian = self.update_val(self.options_vegetarian, rest.get_options_vegetarian(), user_rating.value)
        self.meal_takeout = self.update_val(self.meal_takeout, rest.get_meal_takeout(), user_rating.value)
        self.accessible_wheelchair = self.update_val(self.accessible_wheelchair, rest.get_accessible_wheelchair(), user_rating.value)
        self.seating_outdoor = self.update_val(self.seating_outdoor, rest.get_seating_outdoor(), user_rating.value)


    class Meta:
        unique_together = (('user', 'context'))


def extraInitForUserVector(**kwargs):
    vector = kwargs.get('instance')
    vector.profile = UserProfile.objects.get(user_id=vector.user.id)

post_init.connect(extraInitForUserVector, UserVector)

