from django.shortcuts import render
from django.http import HttpResponse
from django.contrib.auth.decorators import login_required
from django.db.models import Sum
from django.contrib.auth.models import User
from .models import UserProfile, UserRating, Restaurant, UserVector, Context, \
    RestaurantSimilarityMatrix, UserSimilarityMatrix, ValidateRating
from .forms import UserProfileForm,FilterForm, UserRegistrationForm,RecommendForm
from mongoengine import *
from registration.backends.simple.views import RegistrationView
import numpy as np
import heapq
import random


# Create a new class that redirects the user to the index page, if successful at logging
class MyRegistrationView(RegistrationView):
    form_class= UserRegistrationForm
    def get_success_url(self, request, user):
        profile = UserProfile(user_id=user.id)
        profile.save()
        return '/profile'


def index(request):
    return render(request, 'recapp/index.html', {})


@login_required(login_url='/accounts/login/')
def profile(request):
    args = {}
    try:
        instance = request.user.userprofile
    except UserProfile.DoesNotExist:
        instance = None

    if request.method == 'POST':
        form = UserProfileForm(request.POST, instance=instance)
        if form.is_valid():
            profile = form.save(commit=False)
            profile.user_id = request.user.id
            profile.save()
            for context in Context.objects.all():
                user_vector,created = UserVector.objects.get_or_create(user_id = request.user.id, context_id = context.id)
                user_vector.save()
            return render(request, 'recapp/index.html', {})
    else:
        form = UserProfileForm(instance=instance)

    args['form'] = form

    ratings = UserRating.objects.filter(user_id=request.user.id).order_by('-updated_at')
    for rating in ratings:
        rating.restname = Restaurant.objects.get(factual_id = rating.restaurant).name
    args['ratings'] = ratings
    return render(request, 'recapp/profile.html', args)


def toplist(request):
    scores = get_toplist()
    args = {'scores': scores}

    return render(request, 'recapp/toplist.html', args)


@login_required(login_url='/accounts/login/')
def build_profile(request):
    cuisine=None
    price=None
    min_rating = None
    max_rating = None
    posted = False

    if request.method == 'POST': # If the form is submitted
        form = FilterForm(request.POST)
        if form.is_valid():
            cuisine = form.cleaned_data['cuisine']
            if cuisine:
                cuisine = cuisine.name
            price = form.cleaned_data['price']
            if price:
                price = float(price)
            min_rating = float(form.cleaned_data['min_rating'])
            max_rating = float(form.cleaned_data['max_rating'])
            if max_rating == 0.0:
                max_rating = 5.0
            posted = True
    else:
        form = FilterForm()

    final_rests = []
    rests = get_ids(22,cuisine,price,min_rating,max_rating, posted)

    for rest_id in rests:
        rest = Restaurant.objects.get(factual_id=rest_id)
        rest.cuisines = rest.get_cuisine_string()
        rest.priceRange = rest.get_price_range()
        final_rests.append(rest)

    args = {'restaurants': final_rests, 'count': len(final_rests), 'form': form}

    return render(request, 'recapp/build_profile.html', args)


def get_rated_rests(user_id, context_id):
    rated_rests = []
    ratings1 = UserRating.objects.filter(user_id=user_id, context_id=context_id)
    ratings2 = ValidateRating.objects.filter(user_id=user_id, context_id=context_id)
    merged = list(ratings1) + list(ratings2)
    for rating in merged:
        rated_rests.append(rating.restaurant)
    return rated_rests


def get_ids(num=1256, cuisine=None, price=None, min_rating=None, max_rating=None, posted=None):
    if posted:
        ids = set(get_restaurant_ids_from_mongo(cuisine, price, min_rating, max_rating))
    else:
        file = open("recapp/rest_ids.txt","r")
        ids = set([line[:-1] for line in file])
        file.close()

    if num>len(ids):
        num = len(ids)
    return random.sample(ids, num)


@login_required(login_url='/accounts/login/')
def vote(request):
    if request.method == 'GET':
        rest_id = request.GET['rest_id']
        context_id = request.GET.get('context')
        value = float(request.GET['value'])

        if rest_id and vote and context_id:
            user = request.user

            rating = UserRating.objects.get_or_create(restaurant=rest_id, user_id=user.id, context_id=context_id)[0]
            rating.value = value
            rating.save()

            user_vector = UserVector.objects.get_or_create(user_id = request.user.id, context_id = context_id)[0]
            user_vector.update(rating)
            user_vector.save()
    return HttpResponse()


@login_required(login_url='/accounts/login/')
def validation_vote(request):
    if request.method == 'GET':
        context_mapping = {"Business": 1, "Casual": 3, "Romantic": 4, "Special":2}
        rest_id = request.GET['rest_id']
        context_id = context_mapping[request.GET['context']]
        method = request.GET['method']
        value = float(request.GET['value'])

        if rest_id and vote and context_id and method:
            user = request.user

            rating = ValidateRating.objects.get_or_create(restaurant=rest_id, user_id=user.id, context_id=context_id, method=method)[0]
            rating.value = value
            rating.save()

    return HttpResponse()


def get_restaurants(cuisine, price, min_rating, max_rating):
    connect('restaurant')
    if cuisine and price:
        rests = Restaurant.objects(cuisine__contains=cuisine, rating__gte=min_rating, rating__lte=max_rating, price=price)
    elif cuisine and not price:
        rests = Restaurant.objects(cuisine__contains=cuisine, rating__gte=min_rating, rating__lte=max_rating)
    elif not cuisine and price:
        rests = Restaurant.objects(rating__gte=min_rating, rating__lte=max_rating, price=price)
    else:
        rests = Restaurant.objects(rating__gte=min_rating, rating__lte=max_rating)
    return rests


def get_restaurant_ids_from_mongo(cuisine, price, min_rating, max_rating):
    connect('restaurant')
    if cuisine and price:
        rest_objects = Restaurant.objects(cuisine__contains=cuisine, rating__gte=min_rating, rating__lte=max_rating, price=price).only('factual_id')
        rests = [rest_object.factual_id for rest_object in rest_objects]
    elif cuisine and not price:
        rest_objects = Restaurant.objects(cuisine__contains=cuisine, rating__gte=min_rating, rating__lte=max_rating).only('factual_id')
        rests = [rest_object.factual_id for rest_object in rest_objects]
    elif not cuisine and price:
        rest_objects = Restaurant.objects(price=price, rating__gte=min_rating, rating__lte=max_rating).only('factual_id')
        rests = [rest_object.factual_id for rest_object in rest_objects]
    else:
        rest_objects = Restaurant.objects(rating__gte=min_rating, rating__lte=max_rating).only('factual_id')
        rests = [rest_object.factual_id for rest_object in rest_objects]
    return rests


def get_toplist(size=10):
    final_rests = []
    best_romance = UserRating.objects.filter(context=4).values("restaurant").annotate(sum=Sum("value")).order_by('-sum')[:size]
    best_business = UserRating.objects.filter(context=1).values("restaurant").annotate(sum=Sum("value")).order_by('-sum')[:size]
    best_casual = UserRating.objects.filter(context=3).values("restaurant").annotate(sum=Sum("value")).order_by('-sum')[:size]
    best_special = UserRating.objects.filter(context=2).values("restaurant").annotate(sum=Sum("value")).order_by('-sum')[:size]
    best_total = UserRating.objects.values("restaurant").annotate(sum=Sum("value")).order_by('-sum')[:size]
    best_rests = [("business",best_business), ("special",best_special), ("casual",best_casual) ,("romance",best_romance),("total",best_total)]
    cache = {}

    for (context,restSet) in best_rests:
        restList = []
        for dict in restSet:
            id = dict['restaurant']
            rest = cache.get(id, None)

            if not rest:
                rest = Restaurant.objects.get(factual_id=id)
                cache[id] = rest

            score = dict['sum']

            restList.append((rest.name,rest.factual_id,score))

        final_rests.append((context,restList))
    return final_rests


def restaurant(request, factual_id):
    # Create a context dictionary which we can pass to the template rendering engine.
    # We start by containing the name of the category passed by the user.
    context_dict = {'factual_id': factual_id}

    try:
        # Can we find a restuarant with the given name?
        # If we can't, the .get() method raises a DoesNotExist exception.
        # So the .get() method returns one model instance or raises an exception.
        restaurant = Restaurant.objects.get(factual_id=factual_id)
        fields = [('Cuisine', restaurant.get_cuisine_string(),False),
                  ('City', restaurant.locality,False),
                  ('Price', restaurant.get_price_range(),False),
                  ('Rating', restaurant.rating,False),
                  ('Reservations', restaurant.reservations,True),
                  ('Parking', restaurant.parking,True),
                  ('Smoking', restaurant.smoking,True),
                  ('Serves alcohol', restaurant.alcohol,True),
                  ('Wifi', restaurant.wifi,True),
                  ('Vegetarian options', restaurant.options_vegetarian,True),
                  ('Accesible for wheelchairs', restaurant.accessible_wheelchair,True),
                  ('Available for take-out', restaurant.meal_takeout,True),
                  ('Kids menu', restaurant.kids_menu,True),
                  ('Attire',restaurant.attire,False),
                  ('Outdoor seating',restaurant.seating_outdoor,True),
                  ]

        # We add the restaurant object from the database to the context dictionary.
        # We'll use this in the template to verify that the category exists.
        context_dict['restaurant'] = restaurant
        context_dict['fields'] = fields

    except Restaurant.DoesNotExist:
        # We get here if we didn't find the specified category.
        # Don't do anything - the template displays the "no category" message for us.
        pass

    # Go render the response and return it to the client.
    return render(request, 'recapp/restaurant.html', context_dict)


@login_required(login_url='/accounts/login/')
def recommend(request):
    args = {}
    rests = []
    context = None
    if request.method == 'POST':
        form = RecommendForm(request.POST)
        if form.is_valid():
            context = form.cleaned_data['context']

            user_id = request.user.id
            cb,cf,randoms = hybrid(user_id, context.id)
            # create one large list and shuffle
            rests = cb+cf+randoms
            random.shuffle(rests)
    else:
        form = RecommendForm()

    args['form'] = form
    args['rests'] = rests
    args['context'] = context
    return render(request, 'recapp/recommend.html', args)


def special_euclidian_dist(x,y):
    b = x > 0.5
    x_ = b*x + (1-b)*(1-x)
    return np.sqrt(np.sum((x_-y)**2))


def euclidian_dist(x,y):
    return np.sum((x-y)**2)


def cb_rec(user_id, context_id, cold_start, ratings=None):
    if cold_start:
        this_rest = None
        if len(ratings)==0:
            return []
        for rating in ratings:
            if rating.value == 1:
                this_rest = Restaurant.objects.get(factual_id=rating.restaurant)
        if not this_rest:
            return []

        similar_list = RestaurantSimilarityMatrix.objects.get(restaurant=this_rest.factual_id).get_recs()
    else:
        similar_list = UserSimilarityMatrix.objects.get(user=user_id,context=context_id).get_recs()

    results = []
    rated_rests = get_rated_rests(user_id, context_id)
    for rest in similar_list:
        if rest not in rated_rests:
            results.append(rest)
        if(len(results) > 3):
            break
    return results


def cf_rec(user_id, context_id, cold_start, amount=10):
    results = []
    if cold_start:
        similar_users = {}
        try:
            user_vector = np.array(UserVector.objects.get(user_id=user_id,context_id=context_id).get_demo_vector())
        except UserVector.DoesNotExist:
            return results

        for other_user_vector in UserVector.objects.filter(context_id=context_id):
            if user_id == other_user_vector.user_id:
                continue
            similar_users[other_user_vector.user_id] = euclidian_dist(np.asarray(user_vector),np.asarray(other_user_vector.get_demo_vector()))

        user_list = heapq.nsmallest(amount, similar_users, key=similar_users.get)
        most_pop = UserRating.objects.filter(context=context_id,user_id__in=user_list).exclude(user_id=user_id).values("restaurant").annotate(sum=Sum("value")).order_by('-sum')
        for dict in most_pop:
            factual_id = dict['restaurant']
            results.append(factual_id)

    else:
        ratings = build_dict(context_id)
        results = get_cf_recs(ratings,user_id)

    final_results = []
    rated_rests = get_rated_rests(user_id, context_id)
    for rest in results:
        if rest not in rated_rests:
            final_results.append(rest)
        if(len(final_results) > 3):
            break

    return final_results


def build_dict(context_id):
        dict = {}
        for user in User.objects.all():
            ratings = UserRating.objects.filter(user_id=user.id,context_id=context_id)
            ratingDict = {}

            for rating in ratings:
                ratingDict[rating.restaurant] = rating.value

            dict[user.id] = ratingDict
        return dict


def sim_distance(prefs, person1, person2):
    #Get the list of shared_items
    si = False
    for item in prefs[person1]:
        if item in prefs[person2]:
            si = True
            break

    #if they have no rating in common, return 0
    if not si:
        return 0

    #Add up the squares of all differences
    sum_of_squares = sum([pow(prefs[person1][item]-prefs[person2][item],2) for item in prefs[person1] if item in prefs[person2]])

    return 1 / (1 + sum_of_squares)


def get_cf_recs(prefs, person):
    totals = {}

    for other in prefs:
        #don't compare me to myself
        if other == person:
            continue
        sim = sim_distance(prefs,person,other)

        #ignore scores of zero or lower
        if sim <= 0:
            continue
        for item in prefs[other]:
            #only score restaurants i haven't seen yet
            if item not in prefs[person] or prefs[person][item] == 0:
                #Similarity * score
                totals.setdefault(item,0)
                totals[item] += prefs[other][item] * sim

    rankings = sorted(totals, key=totals.get)
    #Return the sorted list
    rankings.reverse()
    return rankings


def get_random_recs(user_id, context_id):
    all_rests = set(get_ids())
    rated_rests = set(get_rated_rests(user_id,context_id))
    rests = list(all_rests-rated_rests)
    random.shuffle(rests)

    results = []
    rated_rests = get_rated_rests(user_id, context_id)
    for rest in rests:
        if rest not in rated_rests:
            results.append(rest)
        if(len(results) > 1):
            break
    return results


def hybrid(user_id, context_id):
    ratings = UserRating.objects.filter(context_id=context_id, user_id=user_id)
    if len(ratings) >= 10:
        cb = cb_rec(user_id, context_id, False)
        cf = cf_rec(user_id,context_id, False)
    else:
        cb = cb_rec(user_id, context_id, True, ratings)
        cf = cf_rec(user_id,context_id, True)

    randoms = get_random_recs(user_id,context_id)

    result_rest_cb = []
    for factual_id in cb[:4]:
        rest = Restaurant.objects.get(factual_id=factual_id)
        rest.cuisines = rest.get_cuisine_string()
        rest.price_range = rest.get_price_range()
        rest.method = "cb"
        result_rest_cb.append(rest)

    result_rest_cf = []
    for factual_id in cf[:4]:
        rest = Restaurant.objects.get(factual_id=factual_id)
        rest.cuisines = rest.get_cuisine_string()
        rest.price_range = rest.get_price_range()
        rest.method="cf"
        result_rest_cf.append(rest)

    result_randoms = []
    for factual_id in randoms[:2]:
        rest = Restaurant.objects.get(factual_id=factual_id)
        rest.cuisines = rest.get_cuisine_string()
        rest.price_range = rest.get_price_range()
        rest.method = "randoms"
        result_randoms.append(rest)

    return result_rest_cb,result_rest_cf,result_randoms



